# Training models built by the model factory

import os
import logging
from functools import partial
from typing import List, Tuple, Iterable, Union

import tensorflow as tf
import keras


import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime

import tensorflow as tf
from ml4h.explorations import predictions_to_pngs

from tensorflow.keras.callbacks import History
from tensorflow.keras.models import Model
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping, ReduceLROnPlateau, Callback

from ml4h.TensorMap import TensorMap

from ml4h.models.diffusion_blocks import DiffusionModel, DiffusionController
from ml4h.plots import plot_metric_history, plot_roc
from ml4h.defines import IMAGE_EXT, MODEL_EXT
from ml4h.models.inspect import plot_and_time_model
from ml4h.models.model_factory import get_custom_objects, make_multimodal_multitask_model
from ml4h.tensor_generators import test_train_valid_tensor_generators, big_batch_from_minibatch_generator


def train_model_from_generators(
    model: Model,
    generate_train: Iterable,
    generate_valid: Iterable,
    training_steps: int,
    validation_steps: int,
    batch_size: int,
    epochs: int,
    patience: int,
    output_folder: str,
    run_id: str,
    inspect_model: bool,
    inspect_show_labels: bool,
    output_tensor_maps: List[TensorMap] = [],
    return_history: bool = False,
    plot: bool = True,
    save_last_model: bool = False,
) -> Union[Model, Tuple[Model, History]]:
    """Train a model from tensor generators for validation and training data.

    Training data lives on disk, it will be loaded by generator functions.
    Plots the metric history after training. Creates a directory to save weights, if necessary.
    Measures runtime and plots architecture diagram if inspect_model is True.
    :param model: The model to optimize
    :param generate_train: Generator function that yields mini-batches of training data.
    :param generate_valid: Generator function that yields mini-batches of validation data.
    :param training_steps: Number of mini-batches in each so-called epoch
    :param validation_steps: Number of validation mini-batches to examine after each epoch.
    :param batch_size: Number of training examples in each mini-batch
    :param epochs: Maximum number of epochs to run regardless of Early Stopping
    :param patience: Number of epochs to wait before reducing learning rate.
    :param output_folder: Directory where output file will be stored
    :param run_id: User-chosen string identifying this run
    :param inspect_model: If True, measure training and inference runtime of the model and generate architecture plot.
    :param inspect_show_labels: If True, show labels on the architecture plot.
    :param output_tensor_maps: List of output TensorMap
    :param return_history: If true return history from training and don't plot the training history
    :param plot: If true, plots the metrics for train and validation set at the end of each epoch
    :param save_last_model: If true saves the model weights from last epoch otherwise saves model with best validation loss

    :return: The optimized model.

    """
    model_file = os.path.join(output_folder, run_id, run_id + MODEL_EXT)
    if not os.path.exists(os.path.dirname(model_file)):
        os.makedirs(os.path.dirname(model_file))

    if inspect_model:
        image_p = os.path.join(output_folder, run_id, 'architecture_graph_' + run_id + IMAGE_EXT)
        plot_and_time_model(model, generate_train, generate_valid, batch_size, training_steps, inspect_show_labels, image_p)

    history = model.fit(
        generate_train, steps_per_epoch=training_steps, epochs=epochs, verbose=1,
        validation_steps=validation_steps, validation_data=generate_valid,
        callbacks=_get_callbacks(patience, model_file, save_last_model),
    )

    logging.info('Model weights saved at: %s' % model_file)
    custom_dict = get_custom_objects(output_tensor_maps)
    model = tf.keras.models.load_model(model_file, custom_objects=custom_dict, compile=False)
    model.compile(optimizer='adam', loss='mse')
    if plot:
        plot_metric_history(history, training_steps, run_id, os.path.dirname(model_file))
    if return_history:
        return model, history
    return model


def _get_callbacks(
    patience: int, model_file: str, save_last_model: bool,
) -> List[Callback]:
    callbacks = [
        EarlyStopping(monitor='val_loss', patience=patience * 3, verbose=1),
        ReduceLROnPlateau(monitor='val_loss', factor=0.5, patience=patience, verbose=1),
        ModelCheckpoint(filepath=model_file, verbose=1, save_best_only=not save_last_model),
    ]

    return callbacks


def train_diffusion_model(args):
    generate_train, generate_valid, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model = DiffusionModel(args.tensor_maps_in[0], args.batch_size, args.dense_blocks, args.block_size, args.conv_x,
                           args.diffusion_loss, args.sigmoid_beta, args.inspect_model)

    model.compile(
        optimizer=tf.optimizers.AdamW(
            learning_rate=args.learning_rate, weight_decay=1e-4,
        ),
        loss=keras.losses.mean_absolute_error if args.diffusion_loss == 'mean_absolute_error' else keras.losses.mean_squared_error,

    )
    batch = next(generate_train)
    for k in batch[0]:
        logging.info(f"input {k} {batch[0][k].shape}")
        feature_batch = batch[0][k]
    for k in batch[1]:
        logging.info(f"label {k} {batch[1][k].shape}")
    checkpoint_path = f"{args.output_folder}{args.id}/{args.id}"

    if os.path.exists(checkpoint_path+'.index'):
        model.load_weights(checkpoint_path)
        logging.info(f'Loaded weights from model checkpoint at: {checkpoint_path}')
    else:
        logging.info(f'No checkpoint at: {checkpoint_path}')

    checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        filepath=checkpoint_path,
        save_weights_only=True,
        monitor="val_i_loss",
        mode="min",
        save_best_only=True,
    )

    callbacks = [checkpoint_callback]

    # calculate mean and variance of training dataset for normalization
    model.normalizer.adapt(feature_batch)

    if args.inspect_model:
        model.network.summary(print_fn=logging.info, expand_nested=True)
        tf.keras.utils.plot_model(
            model.network,
            to_file=f"{args.output_folder}/{args.id}/architecture_diffusion_unet.png",
            show_shapes=True,
            show_dtype=False,
            show_layer_names=True,
            rankdir="TB",
            expand_nested=True,
            dpi=args.dpi,
            layer_range=None,
            show_layer_activations=False,
        )

        prefix_value = f'{args.output_folder}{args.id}/learning_generations/'
        if model.tensor_map.axes() == 2:
            plot_partial = partial(model.plot_ecgs, reseed=args.random_seed, prefix=prefix_value)
        else:
            plot_partial = partial(model.plot_images, reseed=args.random_seed, prefix=prefix_value)
        callbacks.append(keras.callbacks.LambdaCallback(on_epoch_end=plot_partial))

    history = model.fit(
        generate_train,
        steps_per_epoch=args.training_steps,
        epochs=args.epochs,
        validation_data=generate_valid,
        validation_steps=args.validation_steps,
        callbacks=callbacks,
    )
    model.load_weights(checkpoint_path)
    plot_metric_history(history, args.training_steps, args.id, os.path.dirname(checkpoint_path))
    if args.inspect_model:
        metrics = model.evaluate(generate_test, batch_size=args.batch_size, steps=args.test_steps, return_dict=True)
        logging.info(f'Test metrics: {metrics}')

        steps = 1 if args.batch_size > 3 else args.test_steps
        data, labels, paths = big_batch_from_minibatch_generator(generate_test, steps)
        sides = int(np.sqrt(steps*args.batch_size))
        preds = model.plot_reconstructions((data, labels), num_rows=sides, num_cols=sides,
                                           prefix=f'{args.output_folder}/{args.id}/reconstructions/')
        image_out = {args.tensor_maps_in[0].output_name(): data[args.tensor_maps_in[0].input_name()]}
        predictions_to_pngs(preds, args.tensor_maps_in, args.tensor_maps_in, data, image_out, paths, f'{args.output_folder}/{args.id}/reconstructions/')
        if model.tensor_map.axes() == 2:
            model.plot_ecgs(num_rows=4, prefix=os.path.dirname(checkpoint_path))
        else:
            model.plot_images(num_cols=sides, num_rows=sides, prefix=os.path.dirname(checkpoint_path))
    return model


def get_eval_model(args, model_file, output_tmap):
    args.tensor_maps_out = [output_tmap]
    eval_model, _, _, _ = make_multimodal_multitask_model(**args.__dict__)
    return eval_model


def regress_on_batch(diffuser, regressor, controls, tm_out, batch_size):
    control_batch = {tm_out.output_name(): controls}

    control_embed = diffuser.control_embed_model(control_batch)
    generated_images = diffuser.generate(
        control_embed,
        num_images=batch_size,
        diffusion_steps=50,
    )
    control_predictions = regressor.predict(generated_images, verbose=0)
    return control_predictions[:, 0] if tm_out.is_continuous() else control_predictions


def regress_on_controlled_generations(diffuser, regressor, tm_out, batches, batch_size, mean, std, prefix):
    preds = []
    all_controls = []
    # controls = np.arange(-8, 8, 1)

    for i in range(batches):
        if tm_out.is_continuous():
            controls = np.random.normal(mean, std, size=batch_size)
        elif tm_out.is_categorical():
            controls = np.eye(tm_out.shape[-1])[np.random.choice(tm_out.shape[-1], batch_size)]
        preds.extend(regress_on_batch(diffuser, regressor, controls, tm_out, batch_size))
        all_controls.extend(controls)
        if i % 4 == 0:
            logging.info(f'Inferred on {i+1} synthetic diffusion batches of {batches}')

    preds = np.array(preds)
    all_controls = np.array(all_controls)
    logging.info(f'Control Predictions was {preds.shape} Control true was {all_controls.shape}')
    if tm_out.is_continuous():
        preds = preds.flatten()
        all_controls = all_controls.flatten()
        pearson = np.corrcoef(preds, all_controls)[1, 0]
        logging.info(f'Pearson correlation {pearson:0.3f} ')
        plt.scatter(preds, all_controls)
        plt.xlabel(f'Predicted {tm_out.name}')
        plt.ylabel(f'Control {tm_out.name}')
        plt.title(f'''Diffusion Phenotype: {tm_out.name} Control vs Predictions
        Pearson correlation {pearson:0.3f}, $R^2$ {coefficient_of_determination(preds, all_controls):0.3f}, N = {len(preds)}''')
        now_string = datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'scatter_{tm_out.name}_r_{pearson:0.3f}_{now_string}{IMAGE_EXT}')
        os.makedirs(os.path.dirname(figure_path), exist_ok=True)
        plt.savefig(figure_path)
        plt.close()
    elif tm_out.is_categorical():
        plot_roc(preds, all_controls, tm_out.channel_map,
                 f'Diffusion Phenotype: {tm_out.name} Control vs Predictions', prefix)


def interpolate_controlled_generations(diffuser, tensor_maps_out, control_tm, batch_size, prefix):
    control_batch = {}
    if control_tm.is_continuous():
        samples = np.arange(-4, 5, 1)
    elif control_tm.is_categorical():
        samples = np.arange(0, len(control_tm.channel_map), 1)
    num_rows = len(samples)
    num_cols = 4
    plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
    for row, pheno_scale in enumerate(samples):
        for cm in tensor_maps_out:
            if cm == control_tm and control_tm.is_continuous():
                control_batch[cm.output_name()] = np.ones((batch_size,) + cm.shape) * pheno_scale
            elif cm == control_tm and control_tm.is_categorical():
                control_batch[cm.output_name()] = np.zeros((batch_size,) + cm.shape)
                control_batch[cm.output_name()][:, pheno_scale] = 1.0
            else:
                control_batch[cm.output_name()] = np.zeros((batch_size,) + cm.shape)

        control_embed = diffuser.control_embed_model(control_batch)
        generated_images = diffuser.generate(
            control_embed,
            num_images=batch_size,
            diffusion_steps=50,
            reseed=12345,  # hold everything constant except for control signal
        )

        for i_col, col in enumerate(range(num_cols)):
            index = row * num_cols + col
            plt.subplot(num_rows, num_cols, index + 1)
            if len(generated_images.shape) == 3:
                for lead in range(generated_images.shape[-1]):
                    plt.plot(generated_images[i_col, :, lead], label=lead)
            elif len(generated_images.shape) == 4:
                plt.imshow(generated_images[i_col], cmap='gray')
            plt.gca().set_title(f'{control_tm.name[:14]}: {pheno_scale:0.1f}')
            plt.axis("off")

    plt.tight_layout()
    now_string = datetime.now().strftime('%Y-%m-%d_%H-%M')
    figure_path = os.path.join(prefix, f'interpolate_synthetic_{control_tm.name}_{now_string}{IMAGE_EXT}')
    os.makedirs(os.path.dirname(figure_path), exist_ok=True)
    plt.savefig(figure_path)
    plt.close()


def train_diffusion_control_model(args, supervised=False):
    generate_train, generate_valid, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    if supervised:
        supervised_model, _, _, _ = make_multimodal_multitask_model(**args.__dict__)
        model = DiffusionController(
            args.tensor_maps_in[0], args.tensor_maps_out, args.batch_size, args.dense_blocks, args.block_size, args.conv_x,
            args.dense_layers[0], args.attention_window, args.attention_heads, args.attention_modulo, args.diffusion_loss,
            args.sigmoid_beta, args.diffusion_condition_strategy, args.inspect_model, supervised_model, args.supervision_scalar,
        )
    else:
        model = DiffusionController(
            args.tensor_maps_in[0], args.tensor_maps_out, args.batch_size, args.dense_blocks, args.block_size, args.conv_x,
            args.dense_layers[0], args.attention_window, args.attention_heads, args.attention_modulo, args.diffusion_loss,
            args.sigmoid_beta, args.diffusion_condition_strategy, args.inspect_model,
        )

    loss = keras.losses.mean_absolute_error if args.diffusion_loss == 'mean_absolute_error' else keras.losses.mean_squared_error
    model.compile(
        optimizer=tf.optimizers.AdamW(
            learning_rate=args.learning_rate, weight_decay=1e-4,
        ),
        loss=loss,
    )
    batch = next(generate_train)
    for k in batch[0]:
        logging.info(f"input {k} {batch[0][k].shape}")
        feature_batch = batch[0][k]
    for k in batch[1]:
        logging.info(f"label {k} {batch[1][k].shape}")
    checkpoint_path = f"{args.output_folder}{args.id}/{args.id}"
    checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        filepath=checkpoint_path,
        save_weights_only=True,
        monitor="val_i_loss",
        mode="min",
        save_best_only=True,
    )

    callbacks = [checkpoint_callback]


    # calculate mean and variance of training dataset for normalization
    model.normalizer.adapt(feature_batch)
    if args.inspect_model:
        model.network.summary(print_fn=logging.info, expand_nested=True)
        tf.keras.utils.plot_model(
            model.network,
            to_file=f"{args.output_folder}/{args.id}/architecture_{args.id}_unet.png",
            show_shapes=True,
            show_dtype=False,
            show_layer_names=True,
            rankdir="TB",
            expand_nested=True,
            dpi=args.dpi,
            layer_range=None,
            show_layer_activations=False,
        )
        tf.keras.utils.plot_model(
            model.control_embed_model,
            to_file=f"{args.output_folder}/{args.id}/architecture_{args.id}_control_embed.png",
            show_shapes=True,
            show_dtype=False,
            show_layer_names=True,
            rankdir="TB",
            expand_nested=True,
            dpi=args.dpi,
            layer_range=None,
            show_layer_activations=True,
        )
        prefix_value = f'{args.output_folder}{args.id}/learning_generations/'
        # Create a partial function with reseed and prefix pre-filled
        if model.input_map.axes() == 2:
            plot_partial = partial(model.plot_ecgs, reseed=args.random_seed, prefix=prefix_value)
        else:
            plot_partial = partial(model.plot_images, reseed=args.random_seed, prefix=prefix_value)
        callbacks.append(keras.callbacks.LambdaCallback(on_epoch_end=plot_partial))

    if os.path.exists(checkpoint_path+'.index'):
        model.load_weights(checkpoint_path)
        logging.info(f'Loaded weights from model checkpoint at: {checkpoint_path}')
    else:
        logging.info(f'No checkpoint at: {checkpoint_path}')

    history = model.fit(
        generate_train,
        steps_per_epoch=args.training_steps,
        epochs=args.epochs,
        validation_data=generate_valid,
        validation_steps=args.validation_steps,
        callbacks=callbacks,
    )
    plot_metric_history(history, args.training_steps, args.id, os.path.dirname(checkpoint_path))
    model.load_weights(checkpoint_path)

    if args.inspect_model:
        metrics = model.evaluate(generate_test, batch_size=args.batch_size, steps=args.test_steps, return_dict=True)
        logging.info(f'Test metrics: {metrics}')

        steps = 1 if args.batch_size > 3 else args.test_steps
        data, labels, paths = big_batch_from_minibatch_generator(generate_test, steps)
        sides = int(np.sqrt(steps*args.batch_size))
        preds = model.plot_reconstructions((data, labels), num_rows=sides, num_cols=sides,
                                           prefix=f'{args.output_folder}/{args.id}/reconstructions/')

        image_out = {args.tensor_maps_in[0].output_name(): data[args.tensor_maps_in[0].input_name()]}
        predictions_to_pngs(preds, args.tensor_maps_in, args.tensor_maps_in, data, image_out, paths, f'{args.output_folder}/{args.id}/reconstructions/')
        interpolate_controlled_generations(model, args.tensor_maps_out, args.tensor_maps_out[0], args.batch_size,
                                           f'{args.output_folder}/{args.id}/')
        if model.input_map.axes() == 2:
            model.plot_ecgs(num_rows=2, prefix=os.path.dirname(checkpoint_path))
        else:
            model.plot_images(num_cols=sides, num_rows=sides, prefix=os.path.dirname(checkpoint_path))

        for tm_out, model_file in zip(args.tensor_maps_out, args.model_files):
            args.tensor_maps_out = [tm_out]
            args.model_file = model_file
            eval_model, _, _, _ = make_multimodal_multitask_model(**args.__dict__)
            regress_on_controlled_generations(model, eval_model, tm_out, args.test_steps, args.batch_size,
                                              0.0,2.0,f'{args.output_folder}/{args.id}/')

    return model



import argparse
import datetime
import logging

import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions, GoogleCloudOptions, StandardOptions

from ml4h.defines import GCS_BUCKET
from ml4h.tensorize.dataflow import bigquery_ukb_queries


def parse_args():
    now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--id', default=f"run_{now_string}",
        help='User-defined identifier for this pipeline run. '
             'Per Google: the name must consist of only the characters [-a-z0-9], '
             'starting with a letter and ending with a letter or number.',
    )
    parser.add_argument(
        '--tensor_type', default="categorical",
        help='Type of data to be tensorized',
        choices=['categorical', 'continuous', 'icd', 'disease', 'death', 'phecode_disease'],
    )
    parser.add_argument(
        '--bigquery_dataset', default='ukbb_dev',
        help='BigQuery dataset where the data will be drawn from',
    )
    parser.add_argument(
        '--beam_runner', default='DirectRunner',
        help='Apache Beam runner that will execute the pipeline',
        choices=['DirectRunner', 'DataflowRunner'],
    )
    parser.add_argument(
        '--repo_root',
        help='Root directory of the cloned ml repo',
    )
    parser.add_argument(
        '--gcp_project', default='broad-ml4cvd',
        help='Name of the Google Cloud Platform project',
    )
    parser.add_argument(
        '--gcp_region', default='us-central1',
        help='Google Cloud Platform region',
    )
    # parser.add_argument('--gcs_bucket', default='ml4h',
    #                     help='Name of the Google Cloud Storage bucket where tensors will be written to')
    parser.add_argument(
        '--gcs_output_path',
        help='gs:// folder path excluding the bucket name where tensors will be written to '
             'e.g. specifying /path/to/folder will write to gs://<gcs_bucket>/path/to/folder',
    )
    parser.add_argument(
        "--logging_level", default='INFO', help="Logging level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()

    logging.getLogger().setLevel(args.logging_level)

    packaging_args = [
        f'--requirements_file={args.repo_root}/ml4h/tensorize/dataflow/requirements_ml4h_dataflow.txt',
        f'--setup_file={args.repo_root}/setup.py',
    ]

    pipeline_opts = PipelineOptions(flags=packaging_args)
    google_cloud_options = pipeline_opts.view_as(GoogleCloudOptions)
    google_cloud_options.region = args.gcp_region
    google_cloud_options.project = args.gcp_project
    google_cloud_options.job_name = args.id
    google_cloud_options.staging_location = f"gs://{GCS_BUCKET}/dataflow/staging"
    google_cloud_options.temp_location = f"gs://{GCS_BUCKET}/dataflow/temp"
    pipeline_opts.view_as(StandardOptions).runner = args.beam_runner

    pipeline = beam.Pipeline(options=pipeline_opts)

    bigquery_ukb_queries.tensorize_sql_fields(
        pipeline,
        args.gcs_output_path,
        args.bigquery_dataset,
        args.tensor_type,
    )



import datetime
import logging
import os
from typing import Dict, List, Tuple, Sequence
import math

import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt

from tensorflow import keras
from keras import layers

from ml4h.defines import IMAGE_EXT
from ml4h.metrics import KernelInceptionDistance, InceptionScore, MultiScaleSSIM
from ml4h.models.Block import Block
from ml4h.TensorMap import TensorMap


Tensor = tf.Tensor

# sampling
min_signal_rate = 0.05
max_signal_rate = 0.95

# architecture
embedding_dims = 256
embedding_max_frequency = 1000.0

# optimization
ema = 0.999

# plotting
plot_diffusion_steps = 50


def sinusoidal_embedding(x, dims=1):
    embedding_min_frequency = 1.0
    frequencies = tf.exp(
        tf.linspace(
            tf.math.log(embedding_min_frequency),
            tf.math.log(embedding_max_frequency),
            embedding_dims // 2,
        ),
    )
    angular_speeds = 2.0 * math.pi * frequencies
    if dims == 1:
        embeddings = tf.concat([tf.sin(angular_speeds * x)], axis=2)
    elif dims == 2:
        embeddings = tf.concat([tf.sin(angular_speeds * x), tf.cos(angular_speeds * x)], axis=3)
    elif dims == 3:
        embeddings = tf.concat(
            [tf.sin(angular_speeds * x), tf.cos(angular_speeds * x), -tf.sin(angular_speeds * x)],
            axis=4,
        )
    else:
        raise ValueError(f'No support for 4d or more.')
    return embeddings


def residual_block(width, conv, kernel_size, groups=32):
    def apply(x):
        # shortcut
        input_channels = x.shape[-1]
        if input_channels != width:
            residual = conv(width, kernel_size=1)(x)
        else:
            residual = x

        # first GN → SiLU → Conv
        x = tf.keras.layers.GroupNormalization(groups=groups, axis=-1)(x)
        x = layers.Activation('silu')(x)
        x = conv(width, kernel_size=kernel_size, padding="same")(x)

        # second GN → SiLU → Conv
        x = tf.keras.layers.GroupNormalization(groups=groups, axis=-1)(x)
        x = layers.Activation('silu')(x)
        x = conv(width, kernel_size=kernel_size, padding="same")(x)

        # merge
        x = layers.Add()([x, residual])
        return x
    return apply


def down_block(width, block_depth, conv, pool, kernel_size, groups=32):
    def apply(x):
        x, skips = x
        for _ in range(block_depth):
            x = residual_block(width, conv, kernel_size, groups=groups)(x)
            skips.append(x)
        x = pool(pool_size=2)(x)
        return x
    return apply


def up_block(width, block_depth, conv, upsample, kernel_size, groups=32):
    def apply(x):
        x, skips = x
        x = upsample(size=2)(x)
        for _ in range(block_depth):
            x = layers.Concatenate()([x, skips.pop()])
            x = residual_block(width, conv, kernel_size, groups=groups)(x)
        return x
    return apply


def get_network(input_shape, widths, block_depth, kernel_size):
    noisy_images = keras.Input(shape=input_shape)
    conv, upsample, pool, _ = layers_from_shape_control(input_shape)
    noise_variances = keras.Input(shape=[1] * len(input_shape))

    e = layers.Lambda(sinusoidal_embedding)(noise_variances)
    if len(input_shape) == 2:
        e = upsample(size=input_shape[-2])(e)
    else:
        e = upsample(size=input_shape[:-1], interpolation="nearest")(e)

    print(f'e shape: {e.shape} len {len(input_shape)}')
    x = conv(widths[0], kernel_size=1)(noisy_images)
    x = layers.Concatenate()([x, e])

    skips = []
    for width in widths[:-1]:
        x = down_block(width, block_depth, conv, pool, kernel_size)([x, skips])

    for _ in range(block_depth):
        x = residual_block(widths[-1], conv, kernel_size)(x)

    for width in reversed(widths[:-1]):
        x = up_block(width, block_depth, conv, upsample, kernel_size)([x, skips])

    if len(input_shape) > 2:
        x = conv(input_shape[-1], kernel_size=3, padding='same', kernel_initializer="ones")(x)
    x = conv(input_shape[-1], kernel_size=1, kernel_initializer="zeros")(x)

    return keras.Model([noisy_images, noise_variances], x, name="residual_unet")


def layers_from_shape_control(input_shape):
    if len(input_shape) == 2:
        return layers.Conv1D, layers.UpSampling1D, layers.AveragePooling1D, tuple(
            [slice(None), np.newaxis, slice(None)],
        )
    elif len(input_shape) == 3:
        return layers.Conv2D, layers.UpSampling2D, layers.AveragePooling2D, tuple(
            [slice(None), np.newaxis, np.newaxis, slice(None)],
        )
    elif len(input_shape) == 4:
        return layers.Conv3D, layers.UpSampling3D, layers.AveragePooling3D, tuple(
            [slice(None), np.newaxis, np.newaxis, np.newaxis, slice(None)],
        )


def condition_layer_film(input_tensor, control_vector, filters):
    # Transform control into gamma and beta
    gamma = layers.Dense(filters, activation="linear")(control_vector)
    beta = layers.Dense(filters, activation="linear")(control_vector)

    # Reshape gamma and beta to match the spatial dimensions
    # if 4 == len(input_tensor.shape):
    #     gamma = tf.reshape(gamma, (-1, 1, 1, filters))
    #     beta = tf.reshape(beta, (-1, 1, 1, filters))
    # elif 3 == len(input_tensor.shape):
    #     gamma = tf.reshape(gamma, (-1, 1, filters))
    #     beta = tf.reshape(beta, (-1, 1, filters))
    # Apply FiLM (Feature-wise Linear Modulation)
    return input_tensor * gamma + beta


def residual_block_control(
    width,
    conv,
    kernel_size,
    attention_heads,
    condition_strategy,
    groups: int = 32,
):
    def apply(inputs):
        x, control = inputs
        # ─── shortcut ─────────────────────────────────────────────
        in_channels = x.shape[-1]
        if in_channels != width:
            residual = conv(width, kernel_size=1)(x)
        else:
            residual = x

        # ─── conditioning ─────────────────────────────────────────
        if condition_strategy == 'cross_attention':
            x = layers.MultiHeadAttention(
                num_heads=attention_heads, key_dim=width
            )(x, control)
        elif condition_strategy == 'concat':
            x = layers.Concatenate()([x, control])
        elif condition_strategy == 'film':
            x = condition_layer_film(x, control, width)

        # ─── GN → SiLU → Conv → GN → SiLU → Conv → Add ───────────
        x = tf.keras.layers.GroupNormalization(groups=groups, axis=-1)(x)
        x = layers.Activation('silu')(x)
        x = conv(width, kernel_size=kernel_size, padding='same')(x)

        x = tf.keras.layers.GroupNormalization(groups=groups, axis=-1)(x)
        x = layers.Activation('silu')(x)
        x = conv(width, kernel_size=kernel_size, padding='same')(x)

        x = layers.Add()([x, residual])
        return x

    return apply


def down_block_control(
    width,
    block_depth,
    conv,
    pool,
    kernel_size,
    attention_heads,
    condition_strategy,
    groups: int = 32,
):
    def apply(inputs):
        x, skips, control = inputs
        for _ in range(block_depth):
            x = residual_block_control(
                width,
                conv,
                kernel_size,
                attention_heads,
                condition_strategy,
                groups=groups,
            )([x, control])
            skips.append(x)
        x = pool(pool_size=2)(x)
        return x

    return apply


def up_block_control(
    width,
    block_depth,
    conv,
    upsample,
    kernel_size,
    attention_heads,
    condition_strategy,
    groups: int = 32,
):
    def apply(inputs):
        x, skips, control = inputs
        x = upsample(size=2)(x)
        for _ in range(block_depth):
            x = layers.Concatenate()([x, skips.pop()])
            x = residual_block_control(
                width,
                conv,
                kernel_size,
                attention_heads,
                condition_strategy,
                groups=groups,
            )([x, control])
        return x

    return apply


def get_control_network(
    input_shape,
    widths,
    block_depth,
    kernel_size,
    control_size,
    attention_window,
    attention_heads,
    attention_modulo,
    condition_strategy,
    groups: int = 32,
):
    noisy_images = keras.Input(shape=input_shape)
    noise_variances = keras.Input(shape=[1] * len(input_shape))
    conv, upsample, pool, control_idxs = layers_from_shape_control(input_shape)
    control = keras.Input(shape=(control_size,))

    x = conv(widths[0], kernel_size=1)(noisy_images)
    e = layers.Lambda(sinusoidal_embedding)(noise_variances)

    # upsample embeddings & control
    if len(input_shape) == 2:
        e = upsample(size=input_shape[-2])(e)
        c = upsample(size=input_shape[-2])(control[control_idxs])
    else:
        e = upsample(size=input_shape[:-1], interpolation="nearest")(e)
        c = upsample(size=input_shape[:-1])(control[control_idxs])

    x = layers.Concatenate()([x, e])

    skips = []
    # ─── down blocks ────────────────────────────────────────────
    for i, width in enumerate(widths[:-1]):
        use_control = (attention_modulo > 1 and (i+1) % attention_modulo == 0)
        if use_control:
            # recompute c2 at this resolution...
            c2 = upsample(size=x.shape[1:-1] if len(input_shape)>2 else x.shape[-2])(control[control_idxs])
            x = down_block_control(
                width,
                block_depth,
                conv,
                pool,
                kernel_size,
                attention_heads,
                condition_strategy,
                groups=groups,
            )([x, skips, c2])
        else:
            x = down_block(width, block_depth, conv, pool, kernel_size)([x, skips])

    # ─── bottleneck ────────────────────────────────────────────
    c2 = upsample(size=x.shape[1:-1] if len(input_shape)>2 else x.shape[-2])(control[control_idxs])
    for _ in range(block_depth):
        x = residual_block_control(
            widths[-1],
            conv,
            kernel_size,
            attention_heads,
            condition_strategy,
            groups=groups,
        )([x, c2])

    # ─── up blocks ──────────────────────────────────────────────
    for i, width in enumerate(reversed(widths[:-1])):
        use_control = (attention_modulo > 1 and (len(widths)-1 - i) % attention_modulo == 0)
        if use_control:
            up_size = (x.shape[1]*2, x.shape[2]*2) if len(input_shape)==3 else x.shape[-2]*2
            c2 = upsample(size=up_size)(control[control_idxs])
            x = up_block_control(
                width,
                block_depth,
                conv,
                upsample,
                kernel_size,
                attention_heads,
                condition_strategy,
                groups=groups,
            )([x, skips, c2])
        else:
            x = up_block(width, block_depth, conv, upsample, kernel_size)([x, skips])

    # ─── final convs ────────────────────────────────────────────
    if len(input_shape) > 2:
        x = conv(input_shape[-1], kernel_size=3, activation="linear", padding="same", kernel_initializer="ones")(x)
    x = conv(input_shape[-1], kernel_size=1, activation="linear", kernel_initializer="zeros")(x)

    return keras.Model([noisy_images, noise_variances, control], x, name="control_unet")


def get_control_embed_model(output_maps, control_size):
    control_ins = []
    logging.info(f'Build control embedder on :{output_maps}')
    for cm in output_maps:
        control_ins.append(keras.Input(shape=cm.shape, name=cm.output_name()))
    c = layers.Concatenate()(control_ins)
    c = layers.Dense(control_size, activation='linear')(c)
    return keras.Model(control_ins, c, name='control_embed')


class DiffusionModel(keras.Model):
    def __init__(self, tensor_map, batch_size, widths, block_depth, kernel_size, diffusion_loss, sigmoid_beta, inspect_model):
        super().__init__()

        self.tensor_map = tensor_map
        self.batch_size = batch_size
        self.normalizer = layers.Normalization()
        self.network = get_network(self.tensor_map.shape, widths, block_depth, kernel_size)
        self.ema_network = keras.models.clone_model(self.network)
        self.use_sigmoid_loss = diffusion_loss == 'sigmoid'
        self.beta = sigmoid_beta
        self.inspect_model = inspect_model

    def can_apply(self):
        return self.tensor_map.axes() > 1

    def compile(self, **kwargs):
        super().compile(**kwargs)

        self.noise_loss_tracker = keras.metrics.Mean(name="n_loss")
        self.image_loss_tracker = keras.metrics.Mean(name="i_loss")
        self.mse_metric = tf.keras.metrics.MeanSquaredError(name="mse")
        self.mae_metric = tf.keras.metrics.MeanAbsoluteError(name="mae")
        if self.tensor_map.axes() == 3 and self.inspect_model:
            self.kid = KernelInceptionDistance(name = "kid", input_shape = self.tensor_map.shape, kernel_image_size=299)
            self.ms_ssim = MultiScaleSSIM()

    @property
    def metrics(self):
        m = [self.noise_loss_tracker, self.image_loss_tracker, self.mse_metric, self.mae_metric]
        if self.tensor_map.axes() == 3 and self.inspect_model:
            m.append(self.kid)
            m.append(self.ms_ssim)
        return m

    def denormalize(self, images):
        # convert the pixel values back to 0-1 range
        # images = images - tf.math.reduce_mean(images) + images * tf.math.reduce_std(images)
        images = self.normalizer.mean + images * self.normalizer.variance ** 0.5
        # print(f'images max min {images}')
        return images  # tf.clip_by_value(images, 0.0, 1.0)

    def diffusion_schedule(self, diffusion_times):
        # diffusion times -> angles
        start_angle = tf.acos(max_signal_rate)
        end_angle = tf.acos(min_signal_rate)

        diffusion_angles = start_angle + diffusion_times * (end_angle - start_angle)

        # angles -> signal and noise rates
        signal_rates = tf.cos(diffusion_angles)
        noise_rates = tf.sin(diffusion_angles)
        # note that their squared sum is always: sin^2(x) + cos^2(x) = 1

        return noise_rates, signal_rates

    def denoise(self, noisy_images, noise_rates, signal_rates, training):
        # the exponential moving average weights are used at evaluation
        if training:
            network = self.network
        else:
            network = self.ema_network

        # predict noise component and calculate the image component using it
        pred_noises = network([noisy_images, noise_rates ** 2], training=training)
        pred_images = (noisy_images - noise_rates * pred_noises) / signal_rates

        return pred_noises, pred_images

    def reverse_diffusion(self, initial_noise, diffusion_steps):
        # reverse diffusion = sampling
        num_images = initial_noise.shape[0]
        step_size = 1.0 / diffusion_steps

        # important line:
        # at the first sampling step, the "noisy image" is pure noise
        # but its signal rate is assumed to be nonzero (min_signal_rate)
        next_noisy_images = initial_noise
        for step in range(diffusion_steps):
            noisy_images = next_noisy_images

            # separate the current noisy image to its components
            diffusion_times = tf.ones([num_images] + [1] * self.tensor_map.axes()) - step * step_size
            noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
            pred_noises, pred_images = self.denoise(
                noisy_images, noise_rates, signal_rates, training=False,
            )
            # network used in eval mode

            # remix the predicted components using the next signal and noise rates
            next_diffusion_times = diffusion_times - step_size
            next_noise_rates, next_signal_rates = self.diffusion_schedule(
                next_diffusion_times,
            )
            next_noisy_images = (
                    next_signal_rates * pred_images + next_noise_rates * pred_noises
            )
            # this new noisy image will be used in the next step

        return pred_images

    def generate(self, num_images, diffusion_steps, reseed=None):
        # noise -> images -> denormalized images
        if reseed is not None:
            tf.random.set_seed(reseed)
        initial_noise = tf.random.normal(shape=(num_images,) + self.tensor_map.shape)
        generated_images = self.reverse_diffusion(initial_noise, diffusion_steps)
        generated_images = self.denormalize(generated_images)
        return generated_images

    def train_step(self, images_original):
        # normalize images to have standard deviation of 1, like the noises
        images = images_original[0][self.tensor_map.input_name()]
        self.normalizer.update_state(images)
        # images = images['input_lax_4ch_diastole_slice0_224_3d_continuous']
        images = self.normalizer(images, training=True)
        # images = images.numpy() - images.numpy().mean() / images.numpy().std()
        noises = tf.random.normal(shape=(self.batch_size,) + self.tensor_map.shape)

        # sample uniform random diffusion times
        diffusion_times = tf.random.uniform(
            shape=[self.batch_size] + [1] * self.tensor_map.axes(), minval=0.0, maxval=1.0,
        )
        noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
        # mix the images with noises accordingly
        print(f'noises.shape {noises.shape} images.shape {images.shape}')
        noisy_images = signal_rates * images + noise_rates * noises

        with tf.GradientTape() as tape:
            # train the network to separate noisy images to their components
            pred_noises, pred_images = self.denoise(
                noisy_images, noise_rates, signal_rates, training=True,
            )

            noise_loss = self.loss(noises, pred_noises)  # used for training
            image_loss = self.loss(images, pred_images)  # only used as metric
            if self.use_sigmoid_loss:
                signal_rates_squared = tf.square(signal_rates)
                noise_rates_squared = tf.square(noise_rates)

                # Compute log-SNR (lambda_t)
                lambda_t = tf.math.log(signal_rates_squared / noise_rates_squared)
                weight = tf.math.sigmoid(self.beta - lambda_t)
                noise_loss = weight * noise_loss

        gradients = tape.gradient(noise_loss, self.network.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.network.trainable_weights))

        self.noise_loss_tracker.update_state(noise_loss)
        self.image_loss_tracker.update_state(image_loss)
        self.mse_metric.update_state(noises, pred_noises)
        self.mae_metric.update_state(noises, pred_noises)

        # track the exponential moving averages of weights
        for weight, ema_weight in zip(self.network.weights, self.ema_network.weights):
            ema_weight.assign(ema * ema_weight + (1 - ema) * weight)

        # KID is not measured during the training phase for computational efficiency
        return {m.name: m.result() for m in self.metrics}

    def test_step(self, images_original):
        # normalize images to have standard deviation of 1, like the noises
        images = images_original[0][self.tensor_map.input_name()]
        self.normalizer.update_state(images)
        images = self.normalizer(images, training=False)
        # images = images - tf.math.reduce_mean(images) / tf.math.reduce_std(images)
        noises = tf.random.normal(shape=(self.batch_size,) + self.tensor_map.shape)

        # sample uniform random diffusion times
        diffusion_times = tf.random.uniform(
            shape=[self.batch_size] + [1] * self.tensor_map.axes(), minval=0.0, maxval=1.0,
        )
        noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
        # mix the images with noises accordingly
        noisy_images = signal_rates * images + noise_rates * noises

        # use the network to separate noisy images to their components
        pred_noises, pred_images = self.denoise(
            noisy_images, noise_rates, signal_rates, training=False,
        )

        noise_loss = self.loss(noises, pred_noises)
        image_loss = self.loss(images, pred_images)
        if self.use_sigmoid_loss:
            signal_rates_squared = tf.square(signal_rates)
            noise_rates_squared = tf.square(noise_rates)

            # Compute log-SNR (lambda_t)
            lambda_t = tf.math.log(signal_rates_squared / noise_rates_squared)
            weight = tf.math.sigmoid(self.beta - lambda_t)
            noise_loss = weight * noise_loss

        self.image_loss_tracker.update_state(image_loss)
        self.noise_loss_tracker.update_state(noise_loss)
        self.mse_metric.update_state(noises, pred_noises)
        self.mae_metric.update_state(noises, pred_noises)

        # measure KID between real and generated images
        # this is computationally demanding, kid_diffusion_steps has to be small
        if self.tensor_map.axes() == 3 and self.inspect_model:
            images = self.denormalize(images)
            generated_images = self.generate(
                num_images=self.batch_size, diffusion_steps=20
            )
            self.kid.update_state(images, generated_images)
            self.ms_ssim.update_state(images, generated_images, 255)

        return {m.name: m.result() for m in self.metrics}

    def plot_images(self, epoch=None, logs=None, num_rows=1, num_cols=4, reseed=None, prefix='./figures/'):
        # plot random generated images for visual evaluation of generation quality
        generated_images = self.generate(
            num_images=num_rows * num_cols,
            diffusion_steps=plot_diffusion_steps,
            reseed=reseed,
        )

        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                if len(generated_images[index].shape) == 3 and generated_images[index].shape[-1] > 1:
                    plt.imshow(generated_images[index, ..., 0], cmap='gray')  # just plot first frame
                else:
                    plt.imshow(generated_images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'diffusion_image_generations_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()

    def plot_ecgs(self, epoch=None, logs=None, num_rows=1, num_cols=4, reseed=None, prefix='./figures/'):
        # plot random generated images for visual evaluation of generation quality
        generated_images = self.generate(
            num_images=max(self.batch_size, num_rows * num_cols),
            diffusion_steps=plot_diffusion_steps,
            reseed=reseed,
        )

        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                for lead in range(generated_images.shape[-1]):
                    plt.plot(generated_images[index, :, lead], label=lead)
                plt.axis("off")

        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'diffusion_ecg_generations_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()

    def plot_reconstructions(
        self, images_original, diffusion_amount=0, epoch=None, logs=None, num_rows=2, num_cols=2, prefix='./figures/',
    ):
        images = images_original[0][self.tensor_map.input_name()]
        self.normalizer.update_state(images)
        images = self.normalizer(images, training=False)
        noises = tf.random.normal(shape=(num_rows*num_cols,) + self.tensor_map.shape)

        diffusion_times = diffusion_amount * tf.ones(shape=[num_rows*num_cols] + [1] * self.tensor_map.axes())
        noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
        # mix the images with noises accordingly
        noisy_images = signal_rates * images + noise_rates * noises

        # use the network to separate noisy images to their components
        pred_noises, generated_images = self.denoise(
            noisy_images, noise_rates, signal_rates, training=False,
        )
        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(generated_images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'diffusion_reconstructions_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()
        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'input_images_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()
        return generated_images

    def in_paint(self, images_original, masks, diffusion_steps=64, num_rows=3, num_cols=6):
        images = images_original[0][self.tensor_map.input_name()]
        noises = tf.random.normal(shape=(self.batch_size,) + self.tensor_map.shape)
        # reverse diffusion = sampling
        num_images = images.shape[0]
        step_size = 1.0 / max(0.0001, diffusion_steps)

        # important line:
        # at the first sampling step, the "noisy image" is pure noise
        # but its signal rate is assumed to be nonzero (min_signal_rate)
        next_noisy_images = images * masks + noises * (1 - masks)
        for step in range(diffusion_steps):
            noisy_images = next_noisy_images

            # separate the current noisy image to its components
            diffusion_times = tf.ones([num_images] + [1] * self.tensor_map.axes()) - step * step_size
            noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
            pred_noises, generated_images = self.denoise(
                noisy_images, noise_rates, signal_rates, training=False,
            )

            # apply the mask
            generated_images = generated_images * (1 - masks) + images * masks

            # remix the predicted components using the next signal and noise rates
            next_diffusion_times = diffusion_times - step_size
            next_noise_rates, next_signal_rates = self.diffusion_schedule(
                next_diffusion_times,
            )
            next_noisy_images = (
                    next_signal_rates * generated_images + next_noise_rates * pred_noises
            )
            # this new noisy image will be used in the next step

        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(generated_images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        plt.show()
        plt.close()


class DiffusionController(keras.Model):
    def __init__(
        self, tensor_map, output_maps, batch_size, widths, block_depth, conv_x, control_size,
        attention_start, attention_heads, attention_modulo, diffusion_loss, sigmoid_beta, condition_strategy,
        inspect_model, supervisor = None, supervision_scalar = 0.01,
    ):
        super().__init__()

        self.input_map = tensor_map
        self.batch_size = batch_size
        self.output_maps = output_maps
        self.control_embed_model = get_control_embed_model(self.output_maps, control_size)
        self.normalizer = layers.Normalization()
        self.network = get_control_network(self.input_map.shape, widths, block_depth, conv_x, control_size,
                                           attention_start, attention_heads, attention_modulo, condition_strategy)
        self.ema_network = keras.models.clone_model(self.network)
        self.use_sigmoid_loss = diffusion_loss == 'sigmoid'
        self.beta = sigmoid_beta
        self.supervisor = supervisor
        self.supervision_scalar = supervision_scalar
        self.inspect_model = inspect_model


    def compile(self, **kwargs):
        super().compile(**kwargs)
        self.noise_loss_tracker = keras.metrics.Mean(name="n_loss")
        self.image_loss_tracker = keras.metrics.Mean(name="i_loss")
        self.mse_metric = tf.keras.metrics.MeanSquaredError(name="mse")
        self.mae_metric = tf.keras.metrics.MeanAbsoluteError(name="mae")
        if self.supervisor is not None:
            self.supervised_loss_tracker = keras.metrics.Mean(name="supervised_loss")
        if self.input_map.axes() == 3 and self.inspect_model:
            self.kid = KernelInceptionDistance(name = "kid", input_shape = self.input_map.shape, kernel_image_size=299)
            self.ms_ssim = MultiScaleSSIM()

    @property
    def metrics(self):
        m = [self.noise_loss_tracker, self.image_loss_tracker, self.mse_metric, self.mae_metric]
        if self.supervisor is not None:
            m.append(self.supervised_loss_tracker)
        if self.input_map.axes() == 3 and self.inspect_model:
            m.append(self.kid)
            m.append(self.ms_ssim)
        return m

    def denormalize(self, images):
        # convert the pixel values back to 0-1 range
        # images = images - tf.math.reduce_mean(images) + images * tf.math.reduce_std(images)
        images = self.normalizer.mean + images * self.normalizer.variance ** 0.5
        # print(f'images max min {images}')
        return images  # tf.clip_by_value(images, 0.0, 1.0)

    def diffusion_schedule(self, diffusion_times):
        # diffusion times -> angles
        start_angle = tf.acos(max_signal_rate)
        end_angle = tf.acos(min_signal_rate)

        diffusion_angles = start_angle + diffusion_times * (end_angle - start_angle)

        # angles -> signal and noise rates
        signal_rates = tf.cos(diffusion_angles)
        noise_rates = tf.sin(diffusion_angles)
        # note that their squared sum is always: sin^2(x) + cos^2(x) = 1

        return noise_rates, signal_rates

    def denoise(self, control_embed, noisy_images, noise_rates, signal_rates, training):
        # the exponential moving average weights are used at evaluation
        if training:
            network = self.network
        else:
            network = self.ema_network

        # predict noise component and calculate the image component using it
        pred_noises = network([noisy_images, noise_rates ** 2, control_embed], training=training)
        pred_images = (noisy_images - noise_rates * pred_noises) / signal_rates

        return pred_noises, pred_images

    def reverse_diffusion(self, control_embed, initial_noise, diffusion_steps):
        # reverse diffusion = sampling
        num_images = initial_noise.shape[0]
        step_size = 1.0 / diffusion_steps

        # important line:
        # at the first sampling step, the "noisy image" is pure noise
        # but its signal rate is assumed to be nonzero (min_signal_rate)
        next_noisy_images = initial_noise
        pred_images = None
        for step in range(diffusion_steps):
            noisy_images = next_noisy_images

            # separate the current noisy image to its components
            diffusion_times = tf.ones([num_images] + [1] * self.input_map.axes()) - step * step_size
            noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
            pred_noises, pred_images = self.denoise(
                control_embed, noisy_images, noise_rates, signal_rates, training=False,
            )
            # network used in eval mode

            # remix the predicted components using the next signal and noise rates
            next_diffusion_times = diffusion_times - step_size
            next_noise_rates, next_signal_rates = self.diffusion_schedule(
                next_diffusion_times,
            )
            next_noisy_images = (
                    next_signal_rates * pred_images + next_noise_rates * pred_noises
            )
            # this new noisy image will be used in the next step
        return pred_images

    def generate(self, control_embed, num_images, diffusion_steps, reseed=None, renoise=None):
        # noise -> images -> denormalized images

        if reseed is not None:
            tf.random.set_seed(reseed)

        initial_noise = tf.random.normal(shape=(num_images,) + self.input_map.shape)

        generated_images = self.reverse_diffusion(control_embed, initial_noise, diffusion_steps)
        generated_images = self.denormalize(generated_images)
        return generated_images

    def generate_from_noise(self, control_embed, num_images, diffusion_steps, initial_noise):
        generated_images = self.reverse_diffusion(control_embed, initial_noise, diffusion_steps)
        generated_images = self.denormalize(generated_images)
        return generated_images

    def train_step(self, batch):
        # normalize images to have standard deviation of 1, like the noises
        images = batch[0][self.input_map.input_name()]
        self.normalizer.update_state(images)
        images = self.normalizer(images, training=True)

        control_embed = self.control_embed_model(batch[1])

        noises = tf.random.normal(shape=(self.batch_size,) + self.input_map.shape)

        # sample uniform random diffusion times
        diffusion_times = tf.random.uniform(
            shape=[self.batch_size] + [1] * self.input_map.axes(), minval=0.0, maxval=1.0,
        )
        noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
        # mix the images with noises accordingly
        print(f'noises.shape {noises.shape} images.shape {images.shape}')
        noisy_images = signal_rates * images + noise_rates * noises

        with tf.GradientTape(persistent=True) if self.supervisor else tf.GradientTape() as tape:
            # train the network to separate noisy images to their components
            pred_noises, pred_images = self.denoise(
                control_embed, noisy_images, noise_rates, signal_rates, training=True,
            )

            noise_loss = self.loss(noises, pred_noises)  # used for training
            image_loss = self.loss(images, pred_images)  # only used as metric
            if self.use_sigmoid_loss:
                signal_rates_squared = tf.square(signal_rates)
                noise_rates_squared = tf.square(noise_rates)

                # Compute log-SNR (lambda_t)
                lambda_t = tf.math.log(signal_rates_squared / noise_rates_squared)
                weight = tf.math.sigmoid(self.beta - lambda_t)
                noise_loss = weight * noise_loss
            if self.supervisor is not None:
                if self.output_maps[0].is_categorical():
                    loss_fn = tf.keras.losses.CategoricalCrossentropy()
                else:
                    loss_fn = tf.keras.losses.MeanSquaredError()
                supervised_preds = self.supervisor(pred_images, training=True)
                supervised_loss = loss_fn(batch[1][self.output_maps[0].output_name()], supervised_preds)
                self.supervised_loss_tracker.update_state(supervised_loss)
                # Combine losses: add noise_loss and supervised_loss
                noise_loss += self.supervision_scalar * supervised_loss

                # Gradients for self.supervised_model
                supervised_gradients = tape.gradient(supervised_loss, self.supervisor.trainable_weights)
                self.optimizer.apply_gradients(zip(supervised_gradients, self.supervisor.trainable_weights))

        gradients = tape.gradient(noise_loss, self.network.trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, self.network.trainable_weights))

        self.noise_loss_tracker.update_state(noise_loss)
        self.image_loss_tracker.update_state(image_loss)
        self.mse_metric.update_state(noises, pred_noises)
        self.mae_metric.update_state(noises, pred_noises)

        # track the exponential moving averages of weights
        for weight, ema_weight in zip(self.network.weights, self.ema_network.weights):
            ema_weight.assign(ema * ema_weight + (1 - ema) * weight)

        # KID is not measured during the training phase for computational efficiency
        return {m.name: m.result() for m in self.metrics}

    def test_step(self, batch):
        # normalize images to have standard deviation of 1, like the noises
        images = batch[0][self.input_map.input_name()]
        self.normalizer.update_state(images)
        images = self.normalizer(images, training=False)

        control_embed = self.control_embed_model(batch[1])

        noises = tf.random.normal(shape=(self.batch_size,) + self.input_map.shape)

        # sample uniform random diffusion times
        diffusion_times = tf.random.uniform(
            shape=[self.batch_size] + [1] * self.input_map.axes(), minval=0.0, maxval=1.0,
        )
        noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
        # mix the images with noises accordingly
        noisy_images = signal_rates * images + noise_rates * noises

        # use the network to separate noisy images to their components
        pred_noises, pred_images = self.denoise(
            control_embed, noisy_images, noise_rates, signal_rates, training=False,
        )

        noise_loss = self.loss(noises, pred_noises)
        image_loss = self.loss(images, pred_images)
        if self.use_sigmoid_loss:
            signal_rates_squared = tf.square(signal_rates)
            noise_rates_squared = tf.square(noise_rates)

            # Compute log-SNR (lambda_t)
            lambda_t = tf.math.log(signal_rates_squared / noise_rates_squared)
            weight = tf.math.sigmoid(self.beta - lambda_t)
            noise_loss = weight * noise_loss
        if self.supervisor is not None:
            if self.output_maps[0].is_categorical():
                loss_fn = tf.keras.losses.CategoricalCrossentropy()
            else:
                loss_fn = tf.keras.losses.MeanSquaredError()
            supervised_preds = self.supervisor(pred_images, training=True)
            supervised_loss = loss_fn(batch[1][self.output_maps[0].output_name()], supervised_preds)
            self.supervised_loss_tracker.update_state(supervised_loss)
            # Combine losses: add noise_loss and supervised_loss
            noise_loss += self.supervision_scalar*supervised_loss

        self.image_loss_tracker.update_state(image_loss)
        self.noise_loss_tracker.update_state(noise_loss)
        self.mse_metric.update_state(noises, pred_noises)
        self.mae_metric.update_state(noises, pred_noises)

        # measure KID between real and generated images
        # this is computationally demanding, kid_diffusion_steps has to be small
        if self.input_map.axes() == 3 and self.inspect_model:
            images = self.denormalize(images)
            generated_images = self.generate(control_embed,
                num_images=self.batch_size, diffusion_steps=20
            )
            self.kid.update_state(images, generated_images)
            self.ms_ssim.update_state(images, generated_images, 255)

        return {m.name: m.result() for m in self.metrics}


    def plot_images(self, epoch=None, logs=None, num_rows=1, num_cols=4, reseed=None, prefix='./figures/'):
        control_batch = {}
        for cm in self.output_maps:
            control_batch[cm.output_name()] = np.zeros((max(self.batch_size, num_rows * num_cols),) + cm.shape)
            if 'Sex' in cm.name:
                control_batch[cm.output_name()][:, 0] = 1  # all female

        print(f'\nControl batch keys: {list(control_batch.keys())}')
        control_embed = self.control_embed_model(control_batch)
        # plot random generated images for visual evaluation of generation quality
        generated_images = self.generate(
            control_embed,
            num_images=max(self.batch_size, num_rows * num_cols),
            diffusion_steps=plot_diffusion_steps,
            reseed=reseed,
        )
        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(generated_images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'diffusion_image_generations_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()

    def plot_reconstructions(
        self, batch, diffusion_amount=0,
        epoch=None, logs=None, num_rows=4, num_cols=4, prefix='./figures/',
    ):
        images = batch[0][self.input_map.input_name()]
        self.normalizer.update_state(images)
        images = self.normalizer(images, training=False)
        noises = tf.random.normal(shape=(self.batch_size,) + self.input_map.shape)
        diffusion_times = diffusion_amount * tf.ones(shape=[self.batch_size] + [1] * self.input_map.axes())
        noise_rates, signal_rates = self.diffusion_schedule(diffusion_times)
        # mix the images with noises accordingly
        noisy_images = signal_rates * images + noise_rates * noises

        control_embed = self.control_embed_model(batch[1])

        # use the network to separate noisy images to their components
        pred_noises, generated_images = self.denoise(
            control_embed, noisy_images, noise_rates, signal_rates, training=False,
        )
        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(generated_images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'diffusion_image_reconstructions_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()
        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'input_images_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()
        return generated_images


    def control_plot_images(
        self, control_batch, epoch=None, logs=None, num_rows=2, num_cols=8, reseed=None,
        renoise=None,
    ):
        control_embed = self.control_embed_model(control_batch)
        # plot random generated images for visual evaluation of generation quality
        generated_images = self.generate(
            control_embed,
            num_images=max(self.batch_size, num_rows * num_cols),
            diffusion_steps=plot_diffusion_steps,
            reseed=reseed,
            renoise=renoise,
        )

        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(generated_images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        plt.show()
        plt.close()

        return generated_images

    def control_plot_images_noise(self, control_batch, initial_noise, epoch=None, logs=None, num_rows=2, num_cols=8):
        control_embed = self.control_embed_model(control_batch)
        # plot random generated images for visual evaluation of generation quality
        generated_images = self.generate_from_noise(
            control_embed,
            num_images=max(self.batch_size, num_rows * num_cols),
            diffusion_steps=plot_diffusion_steps,
            initial_noise=initial_noise,
        )

        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                plt.imshow(generated_images[index], cmap='gray')
                plt.axis("off")
        plt.tight_layout()
        plt.show()
        plt.close()

        return generated_images

    def control_plot_ecgs(
        self, control_batch, epoch=None, logs=None, num_rows=2, num_cols=8, reseed=None,
        renoise=None,
    ):
        control_embed = self.control_embed_model(control_batch)
        # plot random generated images for visual evaluation of generation quality
        generated_images = self.generate(
            control_embed,
            num_images=max(self.batch_size, num_rows * num_cols),
            diffusion_steps=plot_diffusion_steps,
            reseed=reseed,
            renoise=renoise,
        )

        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                for lead in range(generated_images.shape[-1]):
                    plt.plot(generated_images[index, :, lead], label=lead)
                plt.axis("off")
        plt.tight_layout()
        plt.show()
        plt.close()

        return generated_images

    def plot_ecgs(self, epoch=None, logs=None, num_rows=1, num_cols=4, reseed=None, prefix='./figures/'):
        control_batch = {}
        for cm in self.output_maps:
            control_batch[cm.output_name()] = np.zeros((max(self.batch_size, num_rows * num_cols),) + cm.shape)
            if 'Sex' in cm.name:
                control_batch[cm.output_name()][:, 0] = 1  # all female

        print(f'\nControl batch keys: {list(control_batch.keys())}')
        control_embed = self.control_embed_model(control_batch)

        # plot random generated images for visual evaluation of generation quality
        generated_images = self.generate(
            control_embed,
            num_images=max(self.batch_size, num_rows * num_cols),
            diffusion_steps=plot_diffusion_steps,
            reseed=reseed,
        )
        logging.info(f'Generated ECGs with shape:{generated_images.shape}')
        plt.figure(figsize=(num_cols * 2.0, num_rows * 2.0), dpi=300)
        for row in range(num_rows):
            for col in range(num_cols):
                index = row * num_cols + col
                plt.subplot(num_rows, num_cols, index + 1)
                for lead in range(generated_images.shape[-1]):
                    plt.plot(generated_images[index, :, lead], label=lead)
                plt.axis("off")
        plt.tight_layout()
        now_string = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M')
        figure_path = os.path.join(prefix, f'diffusion_ecg_generations_{now_string}{IMAGE_EXT}')
        if not os.path.exists(os.path.dirname(figure_path)):
            os.makedirs(os.path.dirname(figure_path))
        plt.savefig(figure_path, bbox_inches="tight")
        plt.close()



"""Provides config settings for the logger and a way to load them"""

import sys
import os
import errno
import logging


def load_config(log_level, log_dir, log_file_basename, log_file_suffix):
    from logging import config as logging_config

    try:
        os.makedirs(log_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise e

    logger = logging.getLogger(__name__)

    log_file = "{}/{}_{}.log".format(log_dir, log_file_basename, log_file_suffix)

    try:
        logging_config.dictConfig(_create_config(log_level, log_file))
        success_msg = "Logging configuration was loaded. Log messages can be found at {}.".format(log_file)
        logger.info(success_msg)
    except Exception as e:
        logger.error("Failed to load logging config!")
        raise e


def _create_config(log_level, log_file):
    return {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'simple': {
                'format': '%(asctime)s - %(module)s:%(lineno)d - %(levelname)s - %(message)s',
            },
            'detailed': {
                'format': '%(name)s:%(levelname)s %(module)s:%(lineno)d:  %(message)s',
            },
        },
        'handlers': {
            'console': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': 'simple',
                'stream': sys.stdout,
            },
            'file': {
                'level': log_level,
                'class': 'logging.FileHandler',
                'formatter': 'simple',
                'filename': log_file,
                'mode': 'w',
            },
        },
        'loggers': {
            '': {
                'handlers': ['console', 'file'],
                'level': log_level,
            },
        },
    }
